/************************************************************************
*			     splitdir.c V2.1				*
*		     HP48 Downloaded directory splitter			*
*		     Copyright (C) 1999 Robert Tiismus			*
*		          Robert.Tiismus@mail.ee			*
*		        Last update:  22. Oct. 1999			*
*									*
*  This program is free software; you can redistribute it and/or modify	*
*  it under the terms of the GNU General Public License as published by	*
*  the Free Software Foundation; either version 2 of the License, or	*
*  (at your option) any later version.					*
*									*
*  This program is distributed in the hope that it will be useful,	*
*  but WITHOUT ANY WARRANTY; without even the implied warranty of	*
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
*  GNU General Public License for more details.				*
*									*
*  You should have received a copy of the GNU General Public License	*
*  along with this program; if not, write to the Free Software		*
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		*
************************************************************************/
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <string.h>
#include "hputils.h"

char *RENhp2pc( char * );	// Translate id name to valid DOS name
char *trans2dos( char * );	// Remapping subr for RENhp2pc
char *MAPhp2pc( char * );	// Remaps id to be displayable on PC which
				// will not neccessarily be a valid filename
char dodelim( char * );		// Construct unique delimiter for the
				// informational file ID
void splitRRP( void );		// The actual splitting routine
int  findfile( char *, DIR * );

hpFILE *infile;		// Input file stream
unsigned char extract = 1; // Only look or extract also? Extract by default
unsigned char recurse = 0; // Recurse into subdirectories? Not by default
unsigned char inbackup = 0; // Wrapped into backup object?
int rcounter = -1;	// Recursion counter
char delimiter;		// Filename current delimiter symbol
char *id_name = NULL;	// The ID name of HP48 identifier
char *hp_public;	// ID name after mapping to PC display format
char *OSFname;		// ID name after renaming to valid filename
char startpath[ 201 ];	// Initial working path
char verstr[] = "HPHP48-X";	// Download header for HP48

int main( int argc, char *argv[] )
{ int i, j, l, c;
  unsigned char df = 0;		// Flag for command line parsing
  unsigned char fl = 1;		// Command line parsing
  long obtype;			// Object type number
  char *infile_name = NULL;	// Inputfile name
  char newdir_nam[ 256 ] = "";	// Name of the directory to be created
  char HPversion;		// Version letter of file
  char scratch[ 256 ];
  char s2[ 2 ];

  const char infostr[] = "splitdir V2.1, Copyright (C) 1999 Robert Tiismus";

  puts( infostr );      // Display the informational header

  if( argc < 2 )        // Check for command line parameters
   { fprintf( stderr,
       "Usage:\tsplitdir [options] inputfile\n"
       "-l\t\tlist the contents of the RRP only\n"
       "-r\t\trecurse into subdirectories\n"
       "-d dirname\tname of the base directory for splitting\n" );
     exit( 1 );
   }

  //** Parse the Command-line arguments **
  i = 0;
  while( ++i < argc )
   { if( fl )
      { if( argv[ i ][ 0 ] == '-' )
         { if( strlen( argv[ i ] ) == 1 )
            { fprintf( stderr,
                       "Splitdir ERROR: Invalid command line syntax\n\n" );
              exit( 1 );
            }

           if( !strcmp( argv[ i ], "--" ) )
            { fl = 0;
              continue;
            }

           if( strpbrk( &( argv[ i ][ 1 ] ), "lrd" ) == NULL )
            { fprintf( stderr, "Splitdir ERROR: Invalid option(s): \'%s'\n\n",
                                                                  argv[ i ] );
              exit( 1 );
            }

           scratch[ 0 ] = s2[ 1 ] = '\x0';
           l = strlen( argv[ i ] );
           if( l >= sizeof( scratch ) )
             l = sizeof( scratch ) - 1;
           for( j = 1; j < l; j++ )
             if( strchr( "lrd", s2[ 0 ] = argv[ i ][ j ] ) == NULL )
               strcat( scratch, s2 );
           if( strlen( scratch ) )
            { fprintf( stderr, "Splitdir ERROR: Invalid option(s): \'-%s'\n\n",
                                                                     scratch );
              exit( 1 );
            }

           if( strchr( argv[ i ], 'l' ) != NULL )
             extract = 0;
           if( strchr( argv[ i ], 'r' ) != NULL )
             recurse = 1;
           if( strchr( argv[ i ], 'd' ) != NULL )
             df = 1;
           continue;
         }
      }

     if( strlen( argv[ i ] ) < 255 )
      { if( df )
         { // Name of the directory to be created
           strcpy( newdir_nam, argv[ i ] );
           df = 0;
         }
        else
          // Name of the input file
          infile_name = argv[ i ];
      }
   }

  //** Open the input file **
  if( infile_name == NULL )
   { fprintf( stderr, "Splitdir ERROR: Missing input file name\n" );
     exit( 1 );
   }
  if( ( infile = hp_fopen( infile_name, "rb" ) ) == NULL )
   { perror( infile_name );
     exit( 1 );
   }

  //* Record the current working directory *
  getcwd( startpath, 200 );

  //********************************************************************
  //                   Process the input file
  //********************************************************************

  //** Check for Version string **
  if( ( fgets( verstr, sizeof( verstr ), infile->stream ) == NULL ) ||
      strncmp( verstr, "HPHP48-", 7 ) )
   { fprintf( stderr, "Splitdir ERROR: Not a HP48 file.\n" );
     exit( 1 );
   }
  HPversion = verstr[ 7 ];
  if( HPversion == 'X' )
    puts( "HP48 file version: UNKNOWN\n\n" );
  else
    printf( "HP48 file version: %c\n\n", HPversion );

  //** Check if it is pure RRP or backup, containing RRP **
  if( ( obtype = hpget_a( infile ) ) == DOBAK )
   { hpget_a( infile );                 // skip the size of the backup object
     id_name = hpget_name( infile );    // Read Backup Name
     if( strlen( id_name ) != 0 )       // IF not NULLNAME,
       hpget_b( infile );               // skip the tail char length number
     hp_public = MAPhp2pc( id_name );
     obtype = hpget_a( infile );
     inbackup = 1;
   }

  //** Print additional information about file **
  if( inbackup )
   { c = dodelim( hp_public );
     printf( "Contains %s object in backup %c%s%c\n",
                       obtype_str( obtype ), c, hp_public, c );
   }
  else
    printf( "Contains %s object\n", obtype_str( obtype ) );

  //** Differentiate and prioritize the name categories **
  if( strlen( newdir_nam ) != 0 )
   { OSFname = newdir_nam;
     hp_public = newdir_nam;
   }
  else
    if( id_name != NULL && strlen( id_name ) != 0 )
     { OSFname = RENhp2pc( id_name );
       hp_public = MAPhp2pc( id_name );
     }
    else
     { strcpy( newdir_nam, infile_name );
       strcat( newdir_nam, ".rrp" );
       OSFname = newdir_nam;
       hp_public = newdir_nam;
     }

  if( obtype == DORRP )
    splitRRP();
  else
   { fprintf( stderr, "Splitdir ERROR: Cannot split %s object\n",
                                              obtype_str( obtype ) );
     exit( 1 );
   }

  hp_fclose( infile );
  return 0;
}

//** splitRRP ***********************************
//
//
//***********************************************
void splitRRP()
{ long l;               // Just the plain counter
  int i, c;             // buffer for data, being copied
  long obtype;          // object type number
  long offset,          // contents of the offset field
       last_addr,       // address past last nibble of current object
       link_addr,       // address of the current object link field
       cur_addr,        // address of the current object
       obsize;          // the size of current object
  char dirinf_nam[ 252 ] = "";   // infofile name
  FILE *dirinf;         // infofile stream
  hpFILE *outvar;

  rcounter++;
  if( extract )
   { //* Create new directory and make it current *
     if( strcmp( hp_public, OSFname ) )
      { delimiter = dodelim( hp_public );
        printf( "\nMkdir: %s ( originally %c%s%c )\n",
                                 OSFname, delimiter, hp_public, delimiter );
      }
     else
       printf( "\nMkdir: %s\n", OSFname );

     if( mkdir( OSFname,
                S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH ) == -1 )
      { perror( OSFname );
        exit( 1 );
      }
     if( chdir( OSFname ) == -1 )
      { perror( OSFname );
        exit( 1 );
      }

     // Create the information file
     strcpy( dirinf_nam, OSFname );
     strcat( dirinf_nam, ".inf" );

     if( ( dirinf = fopen( dirinf_nam, "wt" ) ) == NULL )
      { perror( dirinf_nam );
        chdir( startpath );
        exit( 1 );
      }
     if( !rcounter && inbackup )
       fprintf( dirinf, "\t     .Backup\t%s\n", id_name );
   }
  else
    if( !rcounter )
      putchar( '\n' );

  hpget_x( infile );            // skip the attached library number field
  offset = hpget_a( infile );   // get offset to the last ob
  if( !offset )
   { puts( "Empty directory\n" ); // If Empty directory then stop processing
     if( extract )
      { chdir( ".." );
        fclose( dirinf );
      }
     rcounter--;
     return;
   }

  //* seek to the link field of the last ob * (It's 5 nibs ahead of ob)
  hpseek( infile, ( offset - 5 ) - 5, SEEK_CUR );

  cur_addr = hptell( infile );  // Save last object link field address
  hpget_a( infile );            // Skip the link field
  if( strlen( hpget_name( infile ) ) != 0 )     // Skip name
    hpget_b( infile );
  skipob( hpget_a( infile ), infile );  // Skip the current object
  last_addr = hptell( infile );         // Read position past last object
  hpseek( infile, cur_addr, SEEK_SET ); // Recover the position

  //  Start the extracting
  do
   { link_addr = hptell( infile );      // get link field address
     offset = hpget_a( infile );        // get offset to previous entry

     id_name = hpget_name( infile );    // read the current object name
     if( strlen( id_name ) != 0 )
        hpget_b( infile );

     cur_addr = hptell( infile );       // get current position on the stream
     obsize = last_addr - cur_addr;
     obtype = hpget_a( infile );

     hp_public = MAPhp2pc( id_name );
     OSFname = RENhp2pc( id_name );

     if( obtype == DORRP && recurse )
      { if( extract )
         { i = strlen( hp_public );
           delimiter = dodelim( hp_public );
           fprintf( dirinf, "%-12s %c%s%-*c ( %s )\n", OSFname,
                            delimiter, hp_public, 17 - i, delimiter,
                            obtype_str( obtype ) );
         }
        else
          printf( "\n%-15s  <DIR>\n", hp_public );

        splitRRP();
        if( !extract )
         { printf( "Total:\t%6ld%s bytes\n",
                     obsize >> 1,
                     obsize & 1 ? ".5" : "  " );
           if( offset )
             putchar( '\n' );
         }
      }
     else
      { if( extract )
         { if( strcmp( hp_public, OSFname ) )
            { i = strlen( obtype_str( obtype ) );
              printf( "Extracting: %-15s %6ld%s ( %s %-*c as %s\n",
                      hp_public,
                      obsize >> 1,
                      obsize & 1 ? ".5" : "  ",
                      obtype_str( obtype ), 15 - i, ')',
                      OSFname );
            }
           else
             printf( "Extracting: %-15s %6ld%s ( %s )\n",
                     OSFname,
                     obsize >> 1,
                     obsize & 1 ? ".5" : "  ",
                     obtype_str( obtype ) );

           if( ( outvar = hp_fopen( OSFname, "wb" ) ) == NULL )
            { perror( OSFname );
              chdir( startpath );
              exit( 1 );
            }

           /* Start writing */
           fputs( verstr, outvar->stream );	// Version string for uploading
           hpput_a( obtype, outvar ); 	// Object type - has already been read

           // Copy the body of object
           for( l = 5; l < last_addr - cur_addr; l++ )  // Object itself
            { c = getnib( infile );
              putnib( c, outvar );
            }

           hp_fclose( outvar );

           i = strlen( hp_public );
           delimiter = dodelim( hp_public );
           fprintf( dirinf, "%-12s %c%s%-*c ( %s )\n",
                             OSFname,
                             delimiter, hp_public, 17 - i, delimiter,
                             obtype_str( obtype ) );
         }
        else
          printf( "%-15s %6ld%s ( %s )\n",
                   hp_public,
                   obsize >> 1,
                   obsize & 1 ? ".5" : "  ",
                   obtype_str( obtype ) );
      }

     // seek to previous RRW link position
     hpseek( infile, link_addr - offset - 5, SEEK_SET );
     last_addr = link_addr;

   } while( offset );        // continue while there are no more entries

  if( extract )
   { fclose( dirinf );
     chdir( ".." );
   }

  rcounter--;
}

//***********************************************
//
//
//***********************************************
char *RENhp2pc( char *hp_name )
{ DIR *curdir;
  static char hps[ 256 + 1 ];
  int i, l, n;

  strcpy( hps, hp_name );
  trans2dos( hps );

  if( !hps[ 0 ] )
    strcpy( hps, "_" );

  if( ( curdir = opendir( "." ) ) == NULL )
   { perror( "." );
     exit( 1 );
   }
  // If duplicate name - generate new one
  while( !findfile( hps, curdir ) )
   { i = l = strlen( hps );
     while( strchr( "0123456789", hps[ --i ] ) != NULL && i != 0 );
     if( i == l - 1 )
       strcat( hps, ".001" );
     else
      { sscanf( &hps[ i ], "%d", &n );
        n++;
        sprintf( &hps[ i ], "%03d", n );
      }
    }

  closedir( curdir );
  return hps;
}

//****************************************************
// Find out if file named 'name' exists in directory
// refferred by directory stream 'd'
// 0 - exists
// 1 - does not exist
//***************************************************/
int findfile( char *name, DIR *d )
{ struct dirent *dstruct;

  rewinddir( d );
  while( ( dstruct = readdir( d ) ) != NULL )
    if( !strcmp( name, dstruct->d_name ) )
      return 0;

  return -1;
}

//***********************************************
//
//
//***********************************************
char *trans2dos( char *name )
{ int i, l;
  const static char tab4dosnam[] = "#"
  		    "<xvVIS>pd<>#a><v"
  		    "^gdehTlrstwDPO.i";

  l = strlen( name );
  for( i = 0; i < l; i++ )
   { if( ( unsigned char )name[ i ] < ' ' )
       if( ( unsigned char )name[ i ] == ' ' - 1 )
         name[ i ] = '.';
       else
         name[ i ] = '~';
     else
       if( ( unsigned char )name[ i ] >= 0x7F &&
           ( unsigned char )name[ i ] <= 0x9F )
         name[ i ] = tab4dosnam[ ( unsigned char )name[ i ] - 0x7F ];
       else
         if( name[ i ] == '/' )
           name[ i ] = '~';
   }
  return name;
}

//***********************************************
//
//
//***********************************************
char *MAPhp2pc( char *name )
{ int i, l, m;
  static char trname[ 4 * 256 + 1 ];
  const static char tab4hpnam[][ 4 ] =
  { "\\##",
    "\\<)",  "\\x-", "\\.V", "\\v/",  "\\.S", "\\GS", "\\|>", "\\pi",
    "\\.d",  "\\<=", "\\>=", "\\=/",  "\\Ga", "\\->", "\\<-", "\\|v",
    "\\|^",  "\\Gg", "\\Gd", "\\Ge",  "\\Gn", "\\Gh", "\\Gl", "\\Gr",
    "\\Gs",  "\\Gt", "\\Gw", "\\GD",  "\\PI", "\\GW", "\\[]", "\\oo" };

  trname[ 0 ] = 0;
  l = strlen( name );
  for( i = 0; i < l; i++ )
   { if( ( unsigned char )name[ i ] < ' ' )
      { m = strlen( trname );
        sprintf( &trname[ m ], "\\%03d", ( int )name[ i ] );
      }
     else
      { if( ( unsigned char )name[ i ] >= 0x7F &&
            ( unsigned char )name[ i ] <= 0x9F )
          strcat( trname, tab4hpnam[ ( unsigned char )name[ i ] - 0x7F ] );
        else
         { if( name[ i ] == '\\' )
             strcat( trname, "\\\\" );
           else
            { m = strlen( trname );
              trname[ m ] = name[ i ];
              trname[ m + 1 ] = 0;
            }
         }
      }
   }
  return trname;
}

//** dodelim ************************************
//
//
//***********************************************
char dodelim( char name[] )
{ int i, l;
  char delims[] = "'\\/\"!*#$%&+-.:;=?@^_`|~";

  l = sizeof( delims ) - 1;
  for( i = 0; i < l; i++ )
    if( strchr( name, delims[ i ] ) == NULL )
      return delims[ i ];

  for( i = ' ' + 1; i < 256; i++ )
    if( strchr( name, i ) == NULL )
      return ( char )i;

  return ' ';
}
